@@ -1,6 +1,22 @@
 #ifndef COMPAT_H
 #define COMPAT_H
 
+/* _ILP32 and _LP64 are common but not universal, make sure that exactly one
+   of them is defined. */
+#if !defined(_ILP32) && \
+	(SIZEOF_INT == 4) && (SIZEOF_LONG == 4) && (SIZEOF_VOID_P == 4)
+#  define _ILP32
+#endif
+#if !defined(_LP64) && \
+	(SIZEOF_INT == 4) && (SIZEOF_LONG == 8) && (SIZEOF_VOID_P == 8)
+#  define _LP64
+#endif
+#if defined(_ILP32) && defined(_LP64)
+#  error "Cannot have both _ILP32 and _LP64 defined"
+#elif !defined(_ILP32) && !defined(_LP64)
+#  error "Must have one of _ILP32 and _LP64 defined"
+#endif
+
 /* well, this is obviously wrong since it assumes it's 64bit, but older
    GCCs don't define it and we really want it. */
 #ifndef LLONG_MAX
@@ -21,6 +37,17 @@
 #  define NULL ((void *)0)
 #endif
 
+#ifndef __has_extension
+  #define __has_extension(x) 0  // Compatibility with non-clang compilers.
+#endif
+
+#if (defined(__GNUC__) && __GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 5)) || \
+    (defined(__clang__) && (__has_extension(attribute_deprecated_with_message)))
+#  define HAVE_ATTR_DEPRECATED
+int rand(void) __attribute__((deprecated("Do not use rand, use i_rand")));
+int rand_r(unsigned int*) __attribute__((deprecated("Do not use rand_r, use i_rand")));
+#endif
+
 #ifndef __cplusplus
 #ifdef HAVE__BOOL
 typedef _Bool bool;
@@ -61,6 +88,11 @@ typedef unsigned long uint_fast32_t;
 typedef int socklen_t;
 #endif
 
+/* WORDS_BIGENDIAN needs to be undefined if not enabled */
+#if defined(WORDS_BIGENDIAN) && WORDS_BIGENDIAN == 0
+#  undef WORDS_BIGENDIAN
+#endif
+
 #ifdef HAVE_SYS_SYSMACROS_H
 #  include <sys/sysmacros.h>
 #  ifdef HAVE_SYS_MKDEV_H
@@ -78,15 +110,24 @@ typedef int socklen_t;
 #  define ST_ATIME_NSEC(st) ((unsigned long)(st).st_atim.tv_nsec)
 #  define ST_MTIME_NSEC(st) ((unsigned long)(st).st_mtim.tv_nsec)
 #  define ST_CTIME_NSEC(st) ((unsigned long)(st).st_ctim.tv_nsec)
+#  define ST_ATIME_SEC(st) ((unsigned long)(st).st_atim.tv_sec)
+#  define ST_MTIME_SEC(st) ((unsigned long)(st).st_mtim.tv_sec)
+#  define ST_CTIME_SEC(st) ((unsigned long)(st).st_ctim.tv_sec)
 #elif defined (HAVE_STAT_XTIMESPEC)
 #  define HAVE_ST_NSECS
 #  define ST_ATIME_NSEC(st) ((unsigned long)(st).st_atimespec.tv_nsec)
 #  define ST_MTIME_NSEC(st) ((unsigned long)(st).st_mtimespec.tv_nsec)
 #  define ST_CTIME_NSEC(st) ((unsigned long)(st).st_ctimespec.tv_nsec)
+#  define ST_ATIME_SEC(st) ((unsigned long)(st).st_atimespec.tv_sec)
+#  define ST_MTIME_SEC(st) ((unsigned long)(st).st_mtimespec.tv_sec)
+#  define ST_CTIME_SEC(st) ((unsigned long)(st).st_ctimespec.tv_sec)
 #else
 #  define ST_ATIME_NSEC(st) 0UL
 #  define ST_MTIME_NSEC(st) 0UL
 #  define ST_CTIME_NSEC(st) 0UL
+#  define ST_ATIME_SEC(st) 0UL
+#  define ST_MTIME_SEC(st) 0UL
+#  define ST_CTIME_SEC(st) 0UL
 #endif
 
 #ifdef HAVE_ST_NSECS
@@ -212,19 +253,19 @@ int i_my_clock_gettime(int clk_id, struct timespec *tp);
    use our own instead if really needed */
 #define i_toupper(x) ((char) toupper((int) (unsigned char) (x)))
 #define i_tolower(x) ((char) tolower((int) (unsigned char) (x)))
-#define i_isalnum(x) isalnum((int) (unsigned char) (x))
-#define i_isalpha(x) isalpha((int) (unsigned char) (x))
-#define i_isascii(x) isascii((int) (unsigned char) (x))
-#define i_isblank(x) isblank((int) (unsigned char) (x))
-#define i_iscntrl(x) iscntrl((int) (unsigned char) (x))
-#define i_isdigit(x) isdigit((int) (unsigned char) (x))
-#define i_isgraph(x) isgraph((int) (unsigned char) (x))
-#define i_islower(x) islower((int) (unsigned char) (x))
-#define i_isprint(x) isprint((int) (unsigned char) (x))
-#define i_ispunct(x) ispunct((int) (unsigned char) (x))
-#define i_isspace(x) isspace((int) (unsigned char) (x))
-#define i_isupper(x) isupper((int) (unsigned char) (x))
-#define i_isxdigit(x) isxdigit((int) (unsigned char) (x))
+#define i_isalnum(x) (isalnum((int) (unsigned char) (x)) != 0)
+#define i_isalpha(x) (isalpha((int) (unsigned char) (x)) != 0)
+#define i_isascii(x) (isascii((int) (unsigned char) (x)) != 0)
+#define i_isblank(x) (isblank((int) (unsigned char) (x)) != 0)
+#define i_iscntrl(x) (iscntrl((int) (unsigned char) (x)) != 0)
+#define i_isdigit(x) (isdigit((int) (unsigned char) (x)) != 0)
+#define i_isgraph(x) (isgraph((int) (unsigned char) (x)) != 0)
+#define i_islower(x) (islower((int) (unsigned char) (x)) != 0)
+#define i_isprint(x) (isprint((int) (unsigned char) (x)) != 0)
+#define i_ispunct(x) (ispunct((int) (unsigned char) (x)) != 0)
+#define i_isspace(x) (isspace((int) (unsigned char) (x)) != 0)
+#define i_isupper(x) (isupper((int) (unsigned char) (x)) != 0)
+#define i_isxdigit(x) (isxdigit((int) (unsigned char) (x)) != 0)
 
 #ifndef EOVERFLOW
 #  define EOVERFLOW ERANGE
@@ -279,6 +320,10 @@ int fdatasync(int);
 #ifndef IO_BLOCK_SIZE
 #  define IO_BLOCK_SIZE 8192
 #endif
+/* Default size for data blocks transferred over the network */
+#ifndef NET_BLOCK_SIZE
+#  define NET_BLOCK_SIZE (128*1024)
+#endif
 
 #if !defined(PIPE_BUF) && defined(_POSIX_PIPE_BUF)
 #  define PIPE_BUF (8 * _POSIX_PIPE_BUF) /* for HURD */
