@@ -19,6 +19,7 @@ struct passwd_file {
         struct db_passwd_file *db;
 	pool_t pool;
 	int refcount;
+	struct event *event;
 
 	time_t last_sync_time;
 	char *path;
@@ -33,20 +34,21 @@ struct db_passwd_file {
 	struct db_passwd_file *next;
 
 	int refcount;
+	struct event *event;
 
 	char *path;
 	HASH_TABLE(char *, struct passwd_file *) files;
         struct passwd_file *default_file;
 
-	unsigned int vars:1;
-	unsigned int userdb:1;
-	unsigned int userdb_warn_missing:1;
-	unsigned int debug:1;
+	bool vars:1;
+	bool userdb:1;
+	bool userdb_warn_missing:1;
 };
 
-struct passwd_user *
-db_passwd_file_lookup(struct db_passwd_file *db, struct auth_request *request,
-		      const char *username_format);
+int db_passwd_file_lookup(struct db_passwd_file *db,
+			  struct auth_request *request,
+			  const char *username_format,
+			  struct passwd_user **user_r);
 
 struct db_passwd_file *
 db_passwd_file_init(const char *path, bool userdb, bool debug);
