@@ -12,6 +12,7 @@
 
 struct file_lock;
 struct file_create_settings;
+struct fs;
 
 /* Default prefix for indexes */
 #define MAIL_INDEX_PREFIX "dovecot.index"
@@ -23,6 +24,8 @@ struct file_create_settings;
 
 #define MAIL_SHARED_STORAGE_NAME "shared"
 
+#define MAIL_STORAGE_LOST_MAILBOX_PREFIX "recovered-lost-folder-"
+
 enum mail_storage_list_index_rebuild_reason {
 	/* Mailbox list index was found to be corrupted. */
 	MAIL_STORAGE_LIST_INDEX_REBUILD_REASON_CORRUPTED,
@@ -66,13 +69,12 @@ struct mail_storage_vfuncs {
 					 const char *vname,
 					 enum mailbox_flags flags);
 	int (*purge)(struct mail_storage *storage);
-	/* Called when mailbox list index rebuild is requested. Additional
-	   parameter is in storage.list_index_rebuild_reason.
-
+	/* Called when mailbox list index rebuild is requested.
 	   The callback should add any missing mailboxes to the list index.
 	   Returns 0 on success, -1 on temporary failure that didn't properly
 	   rebuild the index. */
-	int (*list_index_corrupted)(struct mail_storage *storage);
+	int (*list_index_rebuild)(struct mail_storage *storage,
+				  enum mail_storage_list_index_rebuild_reason reason);
 };
 
 union mail_storage_module_context {
@@ -107,8 +109,8 @@ enum mail_storage_class_flags {
 	/* Storage deletes all files internally - mailbox list's
 	   delete_mailbox() shouldn't delete anything itself. */
 	MAIL_STORAGE_CLASS_FLAG_NO_LIST_DELETES	= 0x400,
-	/* Storage supports stubs (used for caching purposes). */
-	MAIL_STORAGE_CLASS_FLAG_STUBS = 0x800,
+	/* Storage creates a secondary index */
+	MAIL_STORAGE_CLASS_FLAG_SECONDARY_INDEX	= 0x800,
 };
 
 struct mail_binary_cache {
@@ -137,6 +139,7 @@ struct mail_storage {
 	   MAIL_FETCH_IMAP_BODYSTRUCTURE from the remote server. Adding fields
 	   here avoids adding them to index_mail_data.access_part. */
 	enum mail_fetch_field nonbody_access_fields;
+	struct event_category *event_category;
 
         struct mail_storage_vfuncs v, *vlast;
 
@@ -155,12 +158,16 @@ struct mail_storage {
 	 * uniqueness checking (via strcmp) and never used as a path. */
 	const char *unique_root_dir;
 
+	/* prefix for lost mailbox */
+	const char *lost_mailbox_prefix;
+
 	/* Last error set in mail_storage_set_critical(). */
 	char *last_internal_error;
 
 	char *error_string;
 	enum mail_error error;
 	ARRAY(struct mail_storage_error) error_stack;
+	struct event *event;
 
         const struct mail_storage *storage_class;
 	struct mail_user *user;
@@ -172,21 +179,22 @@ struct mail_storage {
 	struct mail_storage_callbacks callbacks;
 	void *callback_context;
 
-	/* FIXME: remove in v2.3 - should be a parameter to
-	   list_index_corrupted() */
-	enum mail_storage_list_index_rebuild_reason list_index_rebuild_reason;
-
 	struct mail_binary_cache binary_cache;
 	/* Filled lazily by mailbox_attribute_*() when accessing shared
 	   attributes. */
 	struct dict *_shared_attr_dict;
 
+	/* optional fs-api object for accessing mailboxes */
+	struct fs *mailboxes_fs;
+
 	/* Module-specific contexts. See mail_storage_module_id. */
 	ARRAY(union mail_storage_module_context *) module_contexts;
 
 	/* Failed to create shared attribute dict, don't try again */
-	unsigned int shared_attr_dict_failed:1;
-	unsigned int last_error_is_internal:1;
+	bool shared_attr_dict_failed:1;
+	bool last_error_is_internal:1;
+	bool rebuilding_list_index:1;
+	bool rebuild_list_index:1;
 };
 
 struct mail_attachment_part {
@@ -237,23 +245,28 @@ struct mailbox_vfuncs {
 	int (*set_subscribed)(struct mailbox *box, bool set);
 
 	int (*attribute_set)(struct mailbox_transaction_context *t,
-			     enum mail_attribute_type type, const char *key,
+			     enum mail_attribute_type type_flags,
+			     const char *key,
 			     const struct mail_attribute_value *value);
-	int (*attribute_get)(struct mailbox_transaction_context *t,
-			     enum mail_attribute_type type, const char *key,
+	int (*attribute_get)(struct mailbox *box,
+			     enum mail_attribute_type type_flags,
+			     const char *key,
 			     struct mail_attribute_value *value_r);
 	struct mailbox_attribute_iter *
 		(*attribute_iter_init)(struct mailbox *box,
-				       enum mail_attribute_type type,
+				       enum mail_attribute_type type_flags,
 				       const char *prefix);
 	const char *(*attribute_iter_next)(struct mailbox_attribute_iter *iter);
 	int (*attribute_iter_deinit)(struct mailbox_attribute_iter *iter);
 
 	/* Lookup sync extension record and figure out if it mailbox has
-	   changed since. Returns 1 = yes, 0 = no, -1 = error. */
+	   changed since. Returns 1 = yes, 0 = no, -1 = error. if quick==TRUE,
+	   return 1 if it's too costly to find out exactly. The reason_r is
+	   set if 1 is returned. */
 	int (*list_index_has_changed)(struct mailbox *box,
 				      struct mail_index_view *list_view,
-				      uint32_t seq);
+				      uint32_t seq, bool quick,
+				      const char **reason_r);
 	/* Update the sync extension record. */
 	void (*list_index_update_sync)(struct mailbox *box,
 				       struct mail_index_transaction *trans,
@@ -277,7 +290,8 @@ struct mailbox_vfuncs {
 
 	struct mailbox_transaction_context *
 		(*transaction_begin)(struct mailbox *box,
-				     enum mailbox_transaction_flags flags);
+				     enum mailbox_transaction_flags flags,
+				     const char *reason);
 	int (*transaction_commit)(struct mailbox_transaction_context *t,
 				  struct mail_transaction_commit_changes *changes_r);
 	void (*transaction_rollback)(struct mailbox_transaction_context *t);
@@ -298,8 +312,27 @@ struct mailbox_vfuncs {
 	int (*search_deinit)(struct mail_search_context *ctx);
 	bool (*search_next_nonblock)(struct mail_search_context *ctx,
 				     struct mail **mail_r, bool *tryagain_r);
-	/* Internal search function which updates ctx->seq */
+	/* Internally used by the search to find the next mail that potentially
+	   matches the search query. The function performs "quick checks" on
+	   the search query that can be done without initializing the mail
+	   struct. For example it can filter out non-matching mails based on
+	   SEARCH_SEQSET and SEARCH_UIDSET. Once a potentially matching mail is
+	   found, ctx->seq is updated to it.
+
+	   Plugins can use this function to skip over any mails that can't
+	   match the search query. Plugins can also set
+	   mail_search_arg.[non]match_always=TRUE
+	   for search args that it already definitely knows will/won't match,
+	   which prevents the standard search from attempting to match them
+	   again.
+
+	   This is used by the FTS plugin to skip over non-matching mails and
+	   to initialize the [non]match_always=TRUE for the search args that
+	   were matched against the FTS indexes. This prevents the standard
+	   search from having to open any email bodies. */
 	bool (*search_next_update_seq)(struct mail_search_context *ctx);
+	int (*search_next_match_mail)(struct mail_search_context *ctx,
+				      struct mail *mail);
 
 	struct mail_save_context *
 		(*save_alloc)(struct mailbox_transaction_context *t);
@@ -351,12 +384,13 @@ struct mailbox {
 	const char *vname;
 	struct mail_storage *storage;
 	struct mailbox_list *list;
+	struct event *event;
 
 	struct mailbox_vfuncs v, *vlast;
 	/* virtual mailboxes: */
 	const struct virtual_mailbox_vfuncs *virtual_vfuncs;
 /* private: */
-	pool_t pool, metadata_pool;
+	pool_t pool;
 	/* Linked list of all mailboxes in this storage */
 	struct mailbox *prev, *next;
 
@@ -377,8 +411,6 @@ struct mailbox {
 	/* Filled lazily when mailbox is opened, use mailbox_get_index_path()
 	   to access it */
 	const char *_index_path;
-	/* Reason for why mailbox is being accessed or NULL if unknown. */
-	const char *reason;
 
 	/* default vfuncs for new struct mails. */
 	const struct mail_vfuncs *mail_vfuncs;
@@ -393,6 +425,7 @@ struct mailbox {
 	const char *index_prefix;
 	enum mailbox_flags flags;
 	unsigned int transaction_count;
+	unsigned int attribute_iter_count;
 	enum mailbox_feature enabled_features;
 	struct mail_msgpart_partial_cache partial_cache;
 	uint32_t vsize_hdr_ext_id;
@@ -425,47 +458,48 @@ struct mailbox {
 
 	/* When FAST open flag is used, the mailbox isn't actually opened until
 	   it's synced for the first time. */
-	unsigned int opened:1;
+	bool opened:1;
 	/* Mailbox was deleted while we had it open. */
-	unsigned int mailbox_deleted:1;
+	bool mailbox_deleted:1;
 	/* Mailbox is being created */
-	unsigned int creating:1;
+	bool creating:1;
 	/* Mailbox is being deleted */
-	unsigned int deleting:1;
+	bool deleting:1;
 	/* Mailbox is being undeleted */
-	unsigned int mailbox_undeleting:1;
+	bool mailbox_undeleting:1;
 	/* Don't use MAIL_INDEX_SYNC_FLAG_DELETING_INDEX for sync flag */
-	unsigned int delete_sync_check:1;
+	bool delete_sync_check:1;
 	/* Delete mailbox only if it's empty */
-	unsigned int deleting_must_be_empty:1;
+	bool deleting_must_be_empty:1;
 	/* The backend wants to skip checking if there are 0 messages before
 	   calling mailbox_list.delete_mailbox() */
-	unsigned int delete_skip_empty_check:1;
+	bool delete_skip_empty_check:1;
 	/* Mailbox was already marked as deleted within this allocation. */
-	unsigned int marked_deleted:1;
+	bool marked_deleted:1;
 	/* TRUE if this is an INBOX for this user */
-	unsigned int inbox_user:1;
+	bool inbox_user:1;
 	/* TRUE if this is an INBOX for this namespace (user or shared) */
-	unsigned int inbox_any:1;
+	bool inbox_any:1;
 	/* When copying to this mailbox, require that mailbox_copy() uses
 	   mailbox_save_*() to actually save a new physical copy rather than
 	   simply incrementing a reference count (e.g. via hard link) */
-	unsigned int disable_reflink_copy_to:1;
+	bool disable_reflink_copy_to:1;
 	/* Don't allow creating any new keywords */
-	unsigned int disallow_new_keywords:1;
+	bool disallow_new_keywords:1;
 	/* Mailbox has been synced at least once */
-	unsigned int synced:1;
+	bool synced:1;
 	/* Updating cache file is disabled */
-	unsigned int mail_cache_disabled:1;
+	bool mail_cache_disabled:1;
 	/* Update first_saved field to mailbox list index. */
-	unsigned int update_first_saved:1;
+	bool update_first_saved:1;
 	/* mailbox_verify_create_name() only checks for mailbox_verify_name() */
-	unsigned int skip_create_name_restrictions:1;
-	/* v2.2.x API kludge: quick-parameter to list_index_has_changed() */
-	unsigned int list_index_has_changed_quick:1;
+	bool skip_create_name_restrictions:1;
 	/* Using LAYOUT=index and mailbox is being opened with a corrupted
 	   mailbox name. Try to revert to the previously known good name. */
-	unsigned int corrupted_mailbox_name:1;
+	bool corrupted_mailbox_name:1;
+	/* mailbox_open() returned MAIL_ERROR_NOTFOUND because the mailbox
+	   doesn't have the LOOKUP ACL right. */
+	bool acl_no_lookup_right:1;
 };
 
 struct mail_vfuncs {
@@ -475,7 +509,7 @@ struct mail_vfuncs {
 	bool (*set_uid)(struct mail *mail, uint32_t uid);
 	void (*set_uid_cache_updates)(struct mail *mail, bool set);
 	bool (*prefetch)(struct mail *mail);
-	void (*precache)(struct mail *mail);
+	int (*precache)(struct mail *mail);
 	void (*add_temp_wanted_fields)(struct mail *mail,
 				       enum mail_fetch_field fields,
 				       struct mailbox_header_lookup_ctx *headers);
@@ -514,10 +548,7 @@ struct mail_vfuncs {
 
 	int (*get_special)(struct mail *mail, enum mail_fetch_field field,
 			   const char **value_r);
-	/* FIXME: v2.3 API should change this to return -1 on failure.
-	   for now NULL means failure so we don't break backwards
-	   compatibility. */
-	struct mail *(*get_real_mail)(struct mail *mail);
+	int (*get_backend_mail)(struct mail *mail, struct mail **real_mail_r);
 
 	void (*update_flags)(struct mail *mail, enum modify_type modify_type,
 			     enum mail_flags flags);
@@ -528,11 +559,9 @@ struct mail_vfuncs {
 	void (*update_pop3_uidl)(struct mail *mail, const char *uidl);
 	void (*expunge)(struct mail *mail);
 	void (*set_cache_corrupted)(struct mail *mail,
-				    enum mail_fetch_field field);
+				    enum mail_fetch_field field,
+				    const char *reason);
 	int (*istream_opened)(struct mail *mail, struct istream **input);
-	void (*set_cache_corrupted_reason)(struct mail *mail,
-					   enum mail_fetch_field field,
-					   const char *reason);
 };
 
 union mail_module_context {
@@ -549,6 +578,8 @@ struct mail_private {
 	   allows mail_log plugin to log the copy operation using the original
 	   mailbox name. */
 	struct mail *vmail;
+	/* Event is created lazily. Use mail_event() to access it. */
+	struct event *_event;
 
 	uint32_t seq_pvt;
 
@@ -562,6 +593,8 @@ struct mail_private {
 	const char *get_stream_reason;
 
 	bool autoexpunged:1;
+	/* mail created by mailbox_search_*() */
+	bool search_mail:1;
 };
 
 struct mailbox_list_context {
@@ -630,11 +663,7 @@ struct mailbox_transaction_context {
 	/* these statistics are never reset by mail-storage API: */
 	struct mailbox_transaction_stats stats;
 	/* Set to TRUE to update stats_* fields */
-	unsigned int stats_track:1;
-	/* We've done some non-transactional (e.g. dovecot-uidlist updates) */
-	unsigned int nontransactional_changes:1;
-	/* FIXME: v2.3: this should be in attribute_get/set() parameters */
-	unsigned int internal_attribute:1;
+	bool stats_track:1;
 };
 
 union mail_search_module_context {
@@ -661,10 +690,14 @@ struct mail_search_context {
 	uint32_t seq;
 	uint32_t progress_cur, progress_max;
 
+	ARRAY(struct mail *) mails;
+	unsigned int unused_mail_idx;
+	unsigned int max_mails;
+
 	ARRAY(union mail_search_module_context *) module_contexts;
 
-	unsigned int seen_lost_data:1;
-	unsigned int progress_hidden:1;
+	bool seen_lost_data:1;
+	bool progress_hidden:1;
 };
 
 struct mail_save_data {
@@ -676,9 +709,9 @@ struct mail_save_data {
 	time_t received_date, save_date;
 	int received_tz_offset;
 
-	uint32_t uid, stub_seq;
+	uint32_t uid;
 	char *guid, *pop3_uidl, *from_envelope;
-	unsigned int pop3_order;
+	uint32_t pop3_order;
 
 	struct ostream *output;
 	struct mail_save_attachment *attach;
@@ -701,23 +734,21 @@ struct mail_save_context {
 
 	/* mailbox_save_alloc() called, but finish/cancel not.
 	   the same context is usually returned by the backends for reuse. */
-	unsigned int unfinished:1;
+	bool unfinished:1;
 	/* mailbox_save_finish() or mailbox_copy() is being called. */
-	unsigned int finishing:1;
+	bool finishing:1;
 	/* mail was copied or moved using saving (requires:
 	   copying_or_moving==TRUE). */
-	unsigned int copying_via_save:1;
+	bool copying_via_save:1;
 	/* mail is being saved, not copied. However, this is set also with
 	   mailbox_save_using_mail() and then copying_or_moving==TRUE. */
-	unsigned int saving:1;
+	bool saving:1;
 	/* mail is being moved - ignore quota (requires:
 	   copying_or_moving==TRUE && saving==FALSE). */
-	unsigned int moving:1;
+	bool moving:1;
 	/* mail is being copied or moved. However, this is set also with
 	   mailbox_save_using_mail() and then saving==TRUE. */
-	unsigned int copying_or_moving:1;
-	/* dest_mail was set via mailbox_save_set_dest_mail() */
-	unsigned int dest_mail_external:1;
+	bool copying_or_moving:1;
 };
 
 struct mailbox_sync_context {
@@ -743,8 +774,14 @@ extern struct mail_storage_module_register mail_storage_module_register;
 /* Storage's module_id for mail_index. */
 extern struct mail_module_register mail_module_register;
 
+extern struct event_category event_category_storage;
+extern struct event_category event_category_mailbox;
+extern struct event_category event_category_mail;
+
 #define MAIL_STORAGE_CONTEXT(obj) \
 	MODULE_CONTEXT(obj, mail_storage_mail_index_module)
+#define MAIL_STORAGE_CONTEXT_REQUIRE(obj) \
+	MODULE_CONTEXT_REQUIRE(obj, mail_storage_mail_index_module)
 extern MODULE_CONTEXT_DEFINE(mail_storage_mail_index_module,
 			     &mail_index_module_register);
 
@@ -758,6 +795,10 @@ void mail_storage_set_error(struct mail_storage *storage,
 			    enum mail_error error, const char *string);
 void mail_storage_set_critical(struct mail_storage *storage,
 			       const char *fmt, ...) ATTR_FORMAT(2, 3);
+void mailbox_set_critical(struct mailbox *box,
+			  const char *fmt, ...) ATTR_FORMAT(2, 3);
+void mail_set_critical(struct mail *mail,
+		       const char *fmt, ...) ATTR_FORMAT(2, 3);
 void mail_storage_set_internal_error(struct mail_storage *storage);
 void mailbox_set_index_error(struct mailbox *box);
 void mail_storage_set_index_error(struct mail_storage *storage,
@@ -774,6 +815,7 @@ void mail_set_mail_cache_corrupted(struct mail *mail, const char *fmt, ...)
 /* Indicate mail being expunged by autoexpunge */
 void mail_autoexpunge(struct mail *mail);
 
+void mail_event_create(struct mail *mail);
 /* Returns TRUE if everything should already be in memory after this call
    or if prefetching is not supported, i.e. the caller shouldn't do more
    prefetching before this message is handled. */
@@ -788,6 +830,18 @@ bool mail_has_attachment_keywords(struct mail *mail);
    and 1 if attachment was found. */
 int mail_set_attachment_keywords(struct mail *mail);
 
+/* Attempt to start accessing the mail stream. Returns TRUE is ok, FALSE if
+   prevented by mail->lookup_abort. */
+bool mail_stream_access_start(struct mail *mail);
+/* Attempt to start accessing the mail metadata. Returns TRUE is ok, FALSE if
+   prevented by mail->lookup_abort. */
+bool mail_metadata_access_start(struct mail *mail);
+/* Emit mail opened events */
+void mail_opened_event(struct mail *mail);
+
+/* Emit mail expunge_requested event */
+void mail_expunge_requested_event(struct mail *mail);
+
 void mailbox_set_deleted(struct mailbox *box);
 int mailbox_mark_index_deleted(struct mailbox *box, bool del);
 /* Easy wrapper for getting mailbox's MAILBOX_LIST_PATH_TYPE_MAILBOX.
@@ -818,6 +872,8 @@ int mailbox_create_missing_dir(struct mailbox *box,
 			       enum mailbox_list_path_type type);
 /* Returns TRUE if mailbox is autocreated. */
 bool mailbox_is_autocreated(struct mailbox *box);
+/* Returns TRUE if mailbox is autosubscribed. */
+bool mailbox_is_autosubscribed(struct mailbox *box);
 
 /* Returns -1 if error, 0 if failed with EEXIST, 1 if ok */
 int mailbox_create_fd(struct mailbox *box, const char *path, int flags,
@@ -843,4 +899,15 @@ enum mail_index_open_flags
 mail_storage_settings_to_index_flags(const struct mail_storage_settings *set);
 void mailbox_save_context_deinit(struct mail_save_context *ctx);
 
+/* Notify that a sync should be done. */
+void mailbox_sync_notify(struct mailbox *box, uint32_t uid,
+			 enum mailbox_sync_type sync_type);
+
+/* for unit testing */
+int mailbox_verify_name(struct mailbox *box);
+
+int mail_storage_list_index_rebuild_and_set_uncorrupted(struct mail_storage *storage);
+int mail_storage_list_index_rebuild(struct mail_storage *storage,
+				    enum mail_storage_list_index_rebuild_reason reason);
+
 #endif
