@@ -29,10 +29,18 @@
 
 #include <sys/time.h>
 
-#define MAILBOX_LIST_INDEX_HIERARHCY_SEP '~'
+#define MAILBOX_LIST_INDEX_HIERARCHY_SEP '~'
+#define MAILBOX_LIST_INDEX_HIERARCHY_ALT_SEP '^'
 
 #define INDEX_LIST_CONTEXT(obj) \
 	MODULE_CONTEXT(obj, mailbox_list_index_module)
+#define INDEX_LIST_CONTEXT_REQUIRE(obj) \
+	MODULE_CONTEXT_REQUIRE(obj, mailbox_list_index_module)
+
+/* Should the STATUS information for this mailbox not be written to the
+   mailbox list index? */
+#define MAILBOX_IS_NEVER_IN_INDEX(box) \
+	((box)->inbox_any && !(box)->storage->set->mailbox_list_index_include_inbox)
 
 struct mail_index_view;
 struct mailbox_index_vsize;
@@ -85,7 +93,7 @@ struct mailbox_list_index_node {
 	bool corrupted_ext;
 	/* flags are corrupted on disk - need to update it */
 	bool corrupted_flags;
-	const char *name;
+	const char *raw_name;
 };
 
 struct mailbox_list_index {
@@ -112,18 +120,18 @@ struct mailbox_list_index {
 	HASH_TABLE(void *, struct mailbox_list_index_node *) mailbox_hash;
 	struct mailbox_list_index_node *mailbox_tree;
 
-	unsigned int pending_init:1;
-	unsigned int opened:1;
-	unsigned int syncing:1;
-	unsigned int updating_status:1;
-	unsigned int has_backing_store:1;
-	unsigned int index_last_check_changed:1;
-	unsigned int corrupted_names_or_parents:1;
-	unsigned int handling_corruption:1;
-	unsigned int call_corruption_callback:1;
-	unsigned int rebuild_on_missing_inbox:1;
-	unsigned int force_resynced:1;
-	unsigned int force_resync_failed:1;
+	bool pending_init:1;
+	bool opened:1;
+	bool syncing:1;
+	bool updating_status:1;
+	bool has_backing_store:1;
+	bool index_last_check_changed:1;
+	bool corrupted_names_or_parents:1;
+	bool handling_corruption:1;
+	bool call_corruption_callback:1;
+	bool rebuild_on_missing_inbox:1;
+	bool force_resynced:1;
+	bool force_resync_failed:1;
 };
 
 struct mailbox_list_index_iterate_context {
@@ -137,8 +145,8 @@ struct mailbox_list_index_iterate_context {
 	string_t *path;
 	struct mailbox_list_index_node *next_node;
 
-	unsigned int failed:1;
-	unsigned int prefix_inbox_list:1;
+	bool failed:1;
+	bool prefix_inbox_list:1;
 };
 
 extern MODULE_CONTEXT_DEFINE(mailbox_list_index_module,
@@ -154,6 +162,16 @@ void mailbox_list_index_node_get_path(const struct mailbox_list_index_node *node
 void mailbox_list_index_node_unlink(struct mailbox_list_index *ilist,
 				    struct mailbox_list_index_node *node);
 
+/* Return mailbox name encoded into box-name header. */
+const unsigned char *
+mailbox_name_hdr_encode(struct mailbox_list *list, const char *storage_name,
+			size_t *name_len_r);
+/* Return mailbox name decoded from box-name header. */
+const char *
+mailbox_name_hdr_decode_storage_name(struct mailbox_list *list,
+				     const unsigned char *name_hdr,
+				     size_t name_hdr_size);
+
 int mailbox_list_index_index_open(struct mailbox_list *list);
 bool mailbox_list_index_need_refresh(struct mailbox_list_index *ilist,
 				     struct mail_index_view *view);
@@ -166,8 +184,21 @@ void mailbox_list_index_refresh_later(struct mailbox_list *list);
 int mailbox_list_index_handle_corruption(struct mailbox_list *list);
 int mailbox_list_index_set_uncorrupted(struct mailbox_list *list);
 
+/* Returns TRUE and index_r if mailbox list index exists, FALSE if not. */
+bool mailbox_list_index_get_index(struct mailbox_list *list,
+				  struct mail_index **index_r);
+/* Open mailbox list index's view and get the given mailbox's sequence number
+   in it. If require_refreshed is TRUE, the mailbox must have up-to-date
+   information in the mailbox list index. Returns 1 if ok, 0 if mailbox wasn't
+   found or it wasn't up-to-date as requested, -1 if there was an error. The
+   error is stored to the mailbox storage. */
+int mailbox_list_index_view_open(struct mailbox *box, bool require_refreshed,
+				 struct mail_index_view **view_r,
+				 uint32_t *seq_r);
+
 struct mailbox_list_index_node *
-mailbox_list_index_node_find_sibling(struct mailbox_list_index_node *node,
+mailbox_list_index_node_find_sibling(const struct mailbox_list *list,
+				     struct mailbox_list_index_node *node,
 				     const char *name);
 void mailbox_list_index_reset(struct mailbox_list_index *ilist);
 int mailbox_list_index_parse(struct mailbox_list *list,
@@ -186,7 +217,8 @@ bool mailbox_list_index_status(struct mailbox_list *list,
 			       uint32_t seq, enum mailbox_status_items items,
 			       struct mailbox_status *status_r,
 			       uint8_t *mailbox_guid,
-			       struct mailbox_index_vsize *vsize_r);
+			       struct mailbox_index_vsize *vsize_r,
+			       const char **reason_r);
 void mailbox_list_index_status_set_info_flags(struct mailbox *box, uint32_t uid,
 					      enum mailbox_info_flags *flags);
 void mailbox_list_index_update_mailbox_index(struct mailbox *box,
