@@ -12,7 +12,6 @@
 
 #define MBOX_STORAGE_NAME "mbox"
 #define MBOX_SUBSCRIPTION_FILE_NAME ".subscriptions"
-#define MBOX_INDEX_PREFIX "dovecot.index"
 #define MBOX_INDEX_DIR_NAME ".imap"
 #define MBOX_UIDVALIDITY_FNAME "dovecot-uidvalidity"
 
@@ -23,13 +22,19 @@ struct mbox_index_header {
 	uint8_t unused[3];
 	guid_128_t mailbox_guid;
 };
+
+struct mbox_list_index_record {
+	uint32_t mtime;
+	uint32_t size;
+};
+
 struct mbox_storage {
 	struct mail_storage storage;
 
 	const struct mbox_settings *set;
 	enum mbox_lock_type *read_locks;
 	enum mbox_lock_type *write_locks;
-	unsigned int lock_settings_initialized:1;
+	bool lock_settings_initialized:1;
 };
 
 struct mbox_mailbox {
@@ -46,37 +51,49 @@ struct mbox_mailbox {
 	unsigned int mbox_lock_id, mbox_global_lock_id;
 	struct timeout *keep_lock_to;
 	bool mbox_writeonly;
+	unsigned int external_transactions;
 
-	uint32_t mbox_ext_idx, md5hdr_ext_idx;
+	uint32_t mbox_ext_idx, md5hdr_ext_idx, mbox_list_index_ext_id;
 	struct mbox_index_header mbox_hdr;
 	const struct mailbox_update *sync_hdr_update;
 
 	struct mbox_md5_vfuncs md5_v;
 
-	unsigned int no_mbox_file:1;
-	unsigned int invalid_mbox_file:1;
-	unsigned int mbox_broken_offsets:1;
-	unsigned int mbox_save_md5:1;
-	unsigned int mbox_dotlocked:1;
-	unsigned int mbox_used_privileges:1;
-	unsigned int mbox_privileged_locking:1;
-	unsigned int syncing:1;
-	unsigned int backend_readonly:1;
-	unsigned int backend_readonly_set:1;
+	bool no_mbox_file:1;
+	bool invalid_mbox_file:1;
+	bool mbox_broken_offsets:1;
+	bool mbox_save_md5:1;
+	bool mbox_dotlocked:1;
+	bool mbox_used_privileges:1;
+	bool mbox_privileged_locking:1;
+	bool syncing:1;
+	bool backend_readonly:1;
+	bool backend_readonly_set:1;
 };
 
 struct mbox_transaction_context {
 	struct mailbox_transaction_context t;
 	union mail_index_transaction_module_context module_ctx;
 
-	unsigned int mbox_lock_id;
+	unsigned int read_lock_id;
+	unsigned int write_lock_id;
 };
 
+#define MBOX_STORAGE(s)		container_of(s, struct mbox_storage, storage)
+#define MBOX_MAILBOX(s)		container_of(s, struct mbox_mailbox, box)
+#define MBOX_TRANSCTX(s)	container_of(s, struct mbox_transaction_context, t)
+
 extern struct mail_vfuncs mbox_mail_vfuncs;
 extern const char *mbox_hide_headers[], *mbox_save_drop_headers[];
 extern unsigned int mbox_hide_headers_count, mbox_save_drop_headers_count;
 
-int mbox_set_syscall_error(struct mbox_mailbox *mbox, const char *function);
+void mbox_set_syscall_error(struct mbox_mailbox *mbox, const char *function);
+void mbox_istream_set_syscall_error(struct mbox_mailbox *mbox,
+				    struct istream *input,
+				    const char *function);
+void mbox_ostream_set_syscall_error(struct mbox_mailbox *mbox,
+				    struct ostream *output,
+				    const char *function);
 
 struct mailbox_sync_context *
 mbox_storage_sync_init(struct mailbox *box, enum mailbox_sync_flags flags);
